@@ -56,10 +56,11 @@ features:
 =end #'#"#`#
 
 class OptionParser
-  Copyright = %$Copyleft: (c) 1999-2001 Nakada.Nobuyoshi <nobu.nokada@softhome.net> $.freeze
-  RCSID = %w$Id: optparse.rb,v 0.8.4.0 2001/07/09 02:45:57 nobu Exp $[1..-1].each {|s| s.freeze}.freeze
-  Version = RCSID[1].split('.').collect {|s| s.to_i}.extend(Comparable).freeze
-  LastModified = Time.gm(*RCSID[2, 2].join('-').scan(/\d+/).collect {|s| s.to_i})
+  Revision = "0.12"	# -*- package version -*-
+  Copyright = %$Copyleft: (c) 1999-2002 Nakada.Nobuyoshi <nobu.nokada@softhome.net> $.freeze
+  RCSID = %w$Id: optparse.rb,v 0.23 2003/01/31 00:41:04 nobu Exp $[1..-1].each {|s| s.freeze}.freeze
+  Version = (RCSID[1].split('.').collect {|s| s.to_i}.extend(Comparable).freeze if RCSID[1])
+  LastModified = (Time.gm(*RCSID[2, 2].join('-').scan(/\d+/).collect {|s| s.to_i}) if RCSID[2])
   Release = RCSID[2]
 
   NoArgument = [NO_ARGUMENT = :NONE, nil].freeze
@@ -92,10 +93,14 @@ Keyword completion module.
         completion pattern.
 =end #'#"#`#
     def complete(key, pat = nil)
-      pat ||= Regexp.new('\A' + Regexp.quote(key).gsub(/\w+/, '\&\w*'), true)
+      pat ||= Regexp.new('\A' + Regexp.quote(key).gsub(/\w+(?=.)/, '\&\w*'), true)
       canon, sw, k, v = nil
       each do |k, *v|
-	(k.respond_to?(:id2name) ? k.id2name : k) =~ pat or next
+	(if Regexp === k
+	   k === key
+	 else
+	   pat === (defined?(k.id2name) ? k.id2name : k)
+	 end) or next
 	v << k if v.empty?
 	unless canon
 	  canon, sw = k, v
@@ -114,39 +119,21 @@ Keyword completion module.
     Extracts the first element from result of
     ((<OptionParser::Completion#complete>)).
 =end #'#"#`#
-    def convert(opt = nil, *val)
+    def convert(opt = nil, val = nil, *)
       val
     end
   end
 
-=begin private
-== ((:OptionParser::Filling:))
-Extends method (({fill})) like as (({Array#fill})).
-=end #'#"#`#
-  module Filling
-    def fill(val, arg)
-      arg.each {|arg| self[arg] = val}
-      self
-    end
-    def self.extend_object(obj)
-      obj.respond_to?(:fill) or super
-    end
-    def self.append_features(klass)
-      klass.instance_methods.include?('fill') or super
-    end
-  end
-
 =begin private
 == ((:OptionParser::OptionMap:))
 Map from option/keyword string to object with completion.
 === Superclass
 (({Hash}))
 === Including modules
-((<OptionParser::Completion>)), ((<OptionParser::Filling>))
+((<OptionParser::Completion>))
 =end #'#"#`#
   class OptionMap < Hash
     include Completion
-    include Filling
   end
 
 
@@ -168,12 +155,24 @@ Individual switch class.
     def self.guess(arg)
       case arg
       when ""
-        self
-      when /\A[=\s]?\[/
-        Switch::OptionalArgument
+        t = self
+      when /\A=?\[/
+        t = Switch::OptionalArgument
+      when /\A\s+\[/
+        t = Switch::PlacedArgument
       else
-        Switch::RequiredArgument
+        t = Switch::RequiredArgument
       end
+      self >= t or incompatible_argument_styles(arg, t)
+      t
+    end
+
+    def self.incompatible_argument_styles(arg, t)
+      raise ArgumentError, "#{arg}: incompatible argument styles\n  #{self}, #{t}"
+    end
+
+    def self.pattern
+      NilClass
     end
 
 =begin private
@@ -236,8 +235,8 @@ Individual switch class.
 =end #'#"#`#
     def parse(arg, *val)
       if block
-	*val = OptionParser.yieldproc(*val, &conv)
-	return arg, block, *val
+	val = OptionParser.yieldproc(*val, &conv)
+	return arg, block, val
       else
 	return arg, nil
       end
@@ -288,17 +287,8 @@ Individual switch class.
 	l = l.to_s.ljust(width) + ' ' + r if r and !r.empty?
 	yield(indent + l)
       end
-    end
 
-=begin private
---- OptionParser::Switch#-@
-    Instantiates negated boolean switch, which calls ((|@block|)) with
-    inversed value.
-=end #'#"#`#
-    def -@
-      type.new(@pattern, @conv, @short, @long, @arg, @desc) do |val|
-	@block.call !val
-      end
+      self
     end
 
 =begin private
@@ -319,6 +309,11 @@ Switch that takes no arguments.
 	yield(NeedlessArgument, arg) if arg
 	super(arg)
       end
+      def self.incompatible_argument_styles(*)
+      end
+      def self.pattern
+	Object
+      end
     end
 
 =begin private
@@ -338,11 +333,6 @@ Switch that takes an argument.
 	end
 	super(*parse_arg(arg, &error))
       end
-      def self.guess(arg)
-	self >= (t = super) or
-	  raise ArgumentError, "#{arg}: incompatible argument styles\n  #{self}, #{t}"
-	t
-      end
     end
 
 =begin private
@@ -362,10 +352,15 @@ Switch that can omit argument.
 	  super(arg)
 	end
       end
-      def self.guess(arg)
-	self >= (t = super) or
-	  raise ArgumentError, "#{arg}: incompatible argument styles\n  #{self}, #{t}"
-	t
+    end
+
+    class PlacedArgument < self
+      def parse(arg, argv, &error)
+	unless arg
+	  return nil, block, nil if argv.empty? or /\A-/ =~ argv[0]
+	  arg = argv.shift
+	end
+	super(*parse_arg(arg, &error))
       end
     end
   end
@@ -415,7 +410,7 @@ summary feature.
 --- OptionParser::List#reject(type)
     see ((<OptionParser.reject>)).
 =end #'#"#`#
-    def accept(t, pat = nil, &block)
+    def accept(t, pat = /.*/, &block)
       if pat
 	pat.respond_to?(:match) or raise TypeError, "has no `match'"
       else
@@ -445,11 +440,11 @@ summary feature.
       : ((|nlopts|))
         negated long style options list.
 =end #'#"#`#
-    def update(sw, sopts, lopts, nlopts = nil)
+    def update(sw, sopts, lopts, nsw = nil, nlopts = nil)
       o = nil
-      @short.fill(sw, sopts) if sopts and !sopts.empty?
-      @long.fill(sw, lopts) if lopts and !lopts.empty?
-      @long.fill(-sw, nlopts) if nlopts and !nlopts.empty?
+      sopts.each {|o| @short[o] = sw} if sopts
+      lopts.each {|o| @long[o] = sw} if lopts
+      nlopts.each {|o| @long[o] = nsw} if nsw and nlopts
       used = @short.invert.update(@long.invert)
       @list.delete_if {|o| Switch === o and !used[o]}
     end
@@ -591,10 +586,9 @@ The front-end of (({OptionParser})).
     matched to ((<argument pattern>)) is dealed as argument.
 =end #'#"#`#
   ArgumentStyle = {}
-  ArgumentStyle.extend(Filling)
-  ArgumentStyle.fill(Switch::NoArgument, NoArgument)
-  ArgumentStyle.fill(Switch::RequiredArgument, RequiredArgument)
-  ArgumentStyle.fill(Switch::OptionalArgument, OptionalArgument)
+  NoArgument.each {|el| ArgumentStyle[el] = Switch::NoArgument}
+  RequiredArgument.each {|el| ArgumentStyle[el] = Switch::RequiredArgument}
+  OptionalArgument.each {|el| ArgumentStyle[el] = Switch::OptionalArgument}
   ArgumentStyle.freeze
 
 =begin private
@@ -666,7 +660,7 @@ Default options, which never appear in option summary.
     end
   end
   def inc(*args)
-    type.inc(*args)
+    self.class.inc(*args)
   end
 
 =begin
@@ -700,7 +694,7 @@ Default options, which never appear in option summary.
         string pushed back to be first non-option argument
 =end #'#"#`#
   def terminate(arg = nil)
-    type.terminate(arg)
+    self.class.terminate(arg)
   end
   def self.terminate(arg = nil)
     throw :terminate, arg
@@ -778,7 +772,7 @@ Default options, which never appear in option summary.
   attr_accessor :summary_width, :summary_indent
 
   def banner
-    @banner ||= "Usage: #{program_name} [options]\n"
+    @banner ||= "Usage: #{program_name} [options]"
   end
 
   def program_name
@@ -883,7 +877,7 @@ Default options, which never appear in option summary.
 --- OptionParser#to_s
     Returns option summary string.
 =end #'#"#`#
-  def to_str; summarize(banner.to_s.dup) end
+  def to_str; summarize(banner.to_s.sub(/\n?\z/, "\n")) end
   alias to_s to_str
 
 =begin
@@ -955,24 +949,28 @@ Default options, which never appear in option summary.
   end
   private :notwice
 
-  def switch(*opts, &block)
-    short, long, nolong, style, pattern, conv = [], [], []
+  def make_switch(*opts, &block)
+    short, long, nolong, style, pattern, conv, not_pattern, not_conv, not_style = [], [], []
     ldesc, sdesc, desc, arg = [], [], []
     default_style = Switch::NoArgument
     default_pattern = nil
     klass = nil
     o = nil
+    n, q, a = nil
 
     opts.each do |o|
       # argument class
-      next if search(:atype, o) do |default_pattern, conv|
+      next if search(:atype, o) do |pat, c|
 	klass = notwice(o, klass, 'type')
+	if not_style and not_style != Switch::NoArgument
+	  not_pattern, not_conv = pat, c
+	else
+	  default_pattern, conv = pat, c
+	end
       end
-      Module === o and
-	raise ArgumentError, "unsupported argument type: #{o}"
 
       # directly specified pattern(any object possible to match)
-      if o.respond_to?(:match)
+      if !(String === o) and o.respond_to?(:match)
 	pattern = notwice(o, pattern, 'pattern')
 	conv = (pattern.method(:convert).to_proc if pattern.respond_to?(:convert))
 	next
@@ -996,42 +994,68 @@ Default options, which never appear in option summary.
 	else
 	  pattern.update(o)
 	end
+      when Module
+	raise ArgumentError, "unsupported argument type: #{o}"
       when *ArgumentStyle.keys
 	style = notwice(ArgumentStyle[o], style, 'style')
       when /^--no-([^][=\s]*)(.+)?/
-	o = notwice(FalseClass, klass, 'type')
-	default_style = default_style.guess(arg = $2) if $2
-	default_pattern, conv = search(:atype, o)
-	ldesc << "--no-#{$1}"
-	long << 'no-' + (o = $1.downcase)
-	nolong << o
-      when /^--(\[no-\])?([^][=\s]*)(.+)?/
-	o = notwice($3 && !$1 ? Object : TrueClass, klass, 'type')
-	default_style = default_style.guess(arg = $3) if $3
-	default_pattern, conv = search(:atype, o)
-	ldesc << "--#{$1}#{$2}"
-	long << (o = $2.downcase)
-	nolong << 'no-' + o if $1
+	q, a = $1, $2
+	o = notwice(a ? Object : TrueClass, klass, 'type')
+	not_pattern, not_conv = search(:atype, o) unless not_style
+	not_style = (not_style || default_style).guess(arg = a) if a
+	default_style = Switch::NoArgument
+	default_pattern, conv = search(:atype, FalseClass) unless default_pattern
+	ldesc << "--no-#{q}"
+	long << 'no-' + (q = q.downcase)
+	nolong << q
+      when /^--\[no-\]([^][=\s]*)(.+)?/
+	q, a = $1, $2
+	o = notwice(a ? Object : TrueClass, klass, 'type')
+	if a
+	  default_style = default_style.guess(arg = a)
+	  default_pattern, conv = search(:atype, o) unless default_pattern
+	end
+	ldesc << "--#{q}"
+	long << (o = q.downcase)
+	not_pattern, not_conv = search(:atype, FalseClass) unless not_style
+	not_style = Switch::NoArgument
+	nolong << 'no-' + o
+      when /^--([^][=\s]*)(.+)?/
+	q, a = $1, $2
+	if a
+	  o = notwice(NilClass, klass, 'type')
+	  default_style = default_style.guess(arg = a)
+	  default_pattern, conv = search(:atype, o) unless default_pattern
+	end
+	ldesc << "--#{q}"
+	long << (o = q.downcase)
       when /^-(\[\^?\]?(?:[^\\\]]|\\.)*\])(.+)?/
+	q, a = $1, $2
 	o = notwice(Object, klass, 'type')
-	default_style = default_style.guess(arg = $2) if $2
-	default_pattern, conv = search(:atype, o)
-	sdesc << "-#{$1}"
-	short << Regexp.new($1)
+	if a
+	  default_style = default_style.guess(arg = a)
+	  default_pattern, conv = search(:atype, o) unless default_pattern
+	end
+	sdesc << "-#{q}"
+	short << Regexp.new(q)
       when /^-(.)(.+)?/
-	o = notwice(($2 ? Object : TrueClass), klass, 'type')
-	default_style = default_style.guess(arg = $2) if $2
-	default_pattern, conv = search(:atype, o)
-	sdesc << "-#{$1}"
-	short << $1
+	q, a = $1, $2
+	if a
+	  o = notwice(NilClass, klass, 'type')
+	  default_style = default_style.guess(arg = a)
+	  default_pattern, conv = search(:atype, o) unless default_pattern
+	end
+	sdesc << "-#{q}"
+	short << q
       when /^=/
 	style = notwice(default_style.guess(arg = o), style, 'style')
-	default_pattern, conv = search(:atype, Object)
+	default_pattern, conv = search(:atype, Object) unless default_pattern
       else
 	desc.push(o)
       end
     end
 
+    default_pattern, conv = search(:atype, default_style.pattern) unless default_pattern
     s = if short.empty? and long.empty?
 	  raise ArgumentError, "no switch given" if style or pattern or block
 	  desc
@@ -1039,7 +1063,9 @@ Default options, which never appear in option summary.
 	  (style || default_style).new(pattern || default_pattern,
 				       conv, sdesc, ldesc, arg, desc, block)
 	end
-    return s, short, long, nolong
+    return s, short, long,
+      (not_style.new(not_pattern, not_conv, sdesc, ldesc, nil, desc, block) if not_style),
+      nolong
   end
 
 =begin
@@ -1055,23 +1081,39 @@ Default options, which never appear in option summary.
 
     cf. ((<OptionParser#switch>)).
 =end #'#"#`#
+  def define(*opts, &block)
+    top.append(*(sw = make_switch(*opts, &block)))
+    sw[0]
+  end
   def on(*opts, &block)
-    top.append(*switch(*opts, &block))
+    define(*opts, &block)
     self
   end
-  alias def_option on
+  alias def_option define
 
+  def define_head(*opts, &block)
+    top.prepend(*(sw = make_switch(*opts, &block)))
+    sw[0]
+  end
   def on_head(*opts, &block)
-    top.prepend(*switch(*opts, &block))
+    define_head(*opts, &block)
     self
   end
-  alias def_head_option on_head
+  alias def_head_option define_head
 
+  def define_tail(*opts, &block)
+    base.append(*(sw = make_switch(*opts, &block)))
+    sw[0]
+  end
   def on_tail(*opts, &block)
-    base.append(*switch(*opts, &block))
+    define_tail(*opts, &block)
     self
   end
-  alias def_tail_option on_tail
+  alias def_tail_option define_tail
+
+  def separator(string)
+    top.append(string, nil, nil)
+  end
 
 
 =begin
@@ -1094,32 +1136,33 @@ Default options, which never appear in option summary.
   def order(*argv, &block) order!(argv, &block) end
 
   def order!(argv = ARGV, &nonopt)
-    opt, arg, sw, val = nil
+    opt, arg, sw, val, rest = nil
     nonopt ||= proc {|arg| throw :terminate, arg}
     argv.unshift(arg) if arg = catch(:terminate) {
       while arg = argv.shift
 	case arg
 	# long option
 	when /\A--([^=]*)(?:=(.*))?/
+	  opt, rest = $1, $2
 	  begin
-	    sw, = complete(:long, opt = $1)
+	    sw, = complete(:long, opt)
 	  rescue ParseError
 	    raise $!.set_option(arg, true)
 	  end
 	  begin
-	    opt, sw, *val = sw.parse($2, argv) {|*exc| raise(*exc)}
-	    OptionParser.yieldproc(*val, &sw)
+	    opt, sw, val = sw.parse(rest, argv) {|*exc| raise(*exc)}
+	    OptionParser.yieldproc(val, &sw)
 	  rescue ParseError
-	    raise $!.set_option(arg, $2)
+	    raise $!.set_option(arg, rest)
 	  end
 
 	# short option
-	when /\A-(.)((=)(.*)|.+)?/
-	  opt, has_arg, eq, val = $1, $3, $3, $4 || $2
+	when /\A-(.)((=).*|.+)?/
+	  opt, has_arg, eq, val, rest = $1, $3, $3, $2, $2
 	  begin
 	    unless sw = search(:short, opt)
 	      begin
-		sw, = complete(:short, opt, opt)
+		sw, = complete(:short, opt)
 		# short option matched.
 		val = arg.sub(/\A-/, '')
 		has_arg = true
@@ -1127,17 +1170,17 @@ Default options, which never appear in option summary.
 		# if no short options match, try completion with long
 		# options.
 		sw, = complete(:long, opt)
-		eq ||= !$2
+		eq ||= !rest
 	      end
 	    end
 	  rescue ParseError
 	    raise $!.set_option(arg, true)
 	  end
 	  begin
-	    opt, sw, *val = sw.parse(val, argv) {|*exc| raise(*exc) if eq}
+	    opt, sw, val = sw.parse(val, argv) {|*exc| raise(*exc) if eq}
 	    raise InvalidOption, arg if has_arg and !eq and arg == "-#{opt}"
 	    argv.unshift(opt) if opt and (opt = opt.sub(/\A-*/, '-')) != '-'
-	    OptionParser.yieldproc(*val, &sw)
+	    OptionParser.yieldproc(val, &sw)
 	  rescue ParseError
 	    raise $!.set_option(arg, has_arg)
 	  end
@@ -1147,6 +1190,8 @@ Default options, which never appear in option summary.
 	  nonopt.call(arg)
 	end
       end
+
+      nil
     }
 
     argv
@@ -1313,13 +1358,15 @@ Default options, which never appear in option summary.
 : Object
   any string, and no conversion. this is fall-back.
 =end #'#"#`#
-  accept(Object)
+  accept(Object) {|s|s or s.nil?}
+
+  accept(NilClass) {|s|s}
 
 =begin
 : String
   any none-empty string, and no conversion.
 =end #'#"#`#
-  accept(String, /.+/)
+  accept(String, /.+/) {|s,*|s}
 
 =begin
 : Integer
@@ -1327,15 +1374,18 @@ Default options, which never appear in option summary.
   (({0b})), hexadecimal for (({0x})), and decimal for others; with
   optional sign prefix. Converts to (({Integer})).
 =end #'#"#`#
-  octal = '[0-7]+|0b[01]+|0x[\da-f]+'
-  integer = "0#{octal}|[1-9]\\d*"
+  decimal = '\d+(?:_\d+)*'
+  binary = 'b[01]+(?:_[01]+)*'
+  hex = 'x[\da-f]+(?:_[\da-f]+)*'
+  octal = "0(?:[0-7]*(?:_[0-7]+)*|#{binary}|#{hex})"
+  integer = "#{octal}|#{decimal}"
   accept(Integer, %r"\A[-+]?(?:#{integer})"io) {|s| Integer(s) if s}
 
 =begin
 : Float
   Float number format, and converts to (({Float})).
 =end #'#"#`#
-  float = '(?:\d+\.\d*|\.?\d+)(?:E[-+]?\d+)?'
+  float = "(?:#{decimal}(?:\\.(?:#{decimal})?)?|\\.#{decimal})(?:E[-+]?#{decimal})?"
   floatpat = %r"\A[-+]?#{float}"io
   accept(Float, floatpat) {|s| s.to_f if s}
 
@@ -1344,13 +1394,13 @@ Default options, which never appear in option summary.
   Generic numeric format, and converts to (({Integer})) for integer
   format, (({Float})) for float format.
 =end #'#"#`#
-  accept(Numeric, %r"\A[-+]?(?:#{float}|#{integer})"io) {|s| eval(s) if s}
+  accept(Numeric, %r"\A[-+]?(?:#{octal}|#{float})"io) {|s| eval(s) if s}
 
 =begin
 : OptionParser::DecimalInteger
   Decimal integer format, to be converted to (({Integer})).
 =end #'#"#`#
-  DecimalInteger = /\A[-+]?\d+/
+  DecimalInteger = /\A[-+]?#{decimal}/io
   accept(DecimalInteger) {|s| s.to_i if s}
 
 =begin
@@ -1358,7 +1408,7 @@ Default options, which never appear in option summary.
   Ruby/C like octal/hexadecimal/binary integer format, to be converted
   to (({Integer})).
 =end #'#"#`#
-  OctalInteger = %r"\A[-+]?(?:#{octal})"io
+  OctalInteger = /\A[-+]?(?:[0-7]+(?:_[0-7]+)*|0(?:#{binary}|#{hex}))/io
   accept(OctalInteger) {|s| s.oct if s}
 
 =begin
@@ -1377,9 +1427,9 @@ Default options, which never appear in option summary.
 : FalseClass
   Similar to ((<TrueClass>)), but defaulted to (({false})).
 =end #'#"#`#
-  yesno = CompletingHash.new.extend(Filling)
-  yesno.fill(false, %w[- no false])
-  yesno.fill(true, %w[+ yes true])
+  yesno = CompletingHash.new
+  %w[- no false].each {|el| yesno[el] = false}
+  %w[+ yes true].each {|el| yesno[el] = true}
   yesno['nil'] = false		# shoud be nil?
   accept(TrueClass, yesno) {|arg, val| val == nil or val}
   accept(FalseClass, yesno) {|arg, val| val != nil and val}
@@ -1391,7 +1441,6 @@ Default options, which never appear in option summary.
   accept(Array) do |s|
     if s
       s = s.split(',').collect {|s| s unless s.empty?}
-      s.size > 1 or s = [s]	# guard empty or single.
     end
     s
   end
@@ -1447,11 +1496,11 @@ Base class of exceptions from ((<OptionParser>))
     end
 
     def reason
-      @reason || self.type::Reason
+      @reason || self.class::Reason
     end
 
     def inspect
-      '#<' + type.to_s + ': ' + args.join(' ') + '>'
+      '#<' + self.class.to_s + ': ' + args.join(' ') + '>'
     end
 
     def message
